import os
import types
import shutil
import sh  # type: ignore
from tap import Tap
from time import time_ns
from contextlib import contextmanager
from typing import Generator, List, Dict, Any
from pypage import pypage  # type: ignore
from .fs_crawl import (
    FileNode,
    DirNode,
    displayDir,
    NameRegistry,
    fs_crawl,
    Md,
    Page,
    NonMd,
    readfile,
    config_py_file,
    Fore,
    Style,
)


class Args(Tap):  # pyre-ignore[13]
    content: str  # Directory to read the input content from.
    output: str  # Directory to send the output. WARNING: This will be deleted.
    copy_assets: bool = False  # Copy assets instead of symlinking to them
    trailing_slash: bool = (
        False  # Include a trailing slash for links to markdown page directories
    )


class Content(object):
    def __init__(
        self, args: Args, rootDir: DirNode, nameRegistry: NameRegistry
    ) -> None:
        self.args = args
        self.rootDir: DirNode = rootDir
        self.nameRegistry = nameRegistry

    def link(self, srcFile: FileNode, name: str) -> str:
        dstFile: FileNode = self.nameRegistry.lookup(name)
        dstFile.makePublic()  # FixMe: Circular links can make pages public.

        dstFileName = dstFile.fileName
        if isinstance(dstFile.page, NonMd):
            dstFileName = dstFile.page.rectifiedFileName
        elif isinstance(dstFile.page, Md):
            dstFileName = (
                dstFile.pageName
                # Add a "/" trailing slash if arg requests it
                + ("/" if self.args.trailing_slash else "")
            )

        srcPath = self.splitPath(srcFile.fullPath)[:-1]
        dstPath = self.splitPath(dstFile.fullPath)[:-1]
        commonLevel = 0
        for i in range(min(len(srcPath), len(dstPath))):
            if srcPath[i] == dstPath[i]:
                commonLevel += 1
        remainingPath = dstPath[commonLevel:] + [dstFileName]

        relativePath: List[str] = []
        if commonLevel < len(srcPath):
            stepsDown = len(srcPath) - commonLevel
            for _ in range(stepsDown):
                relativePath.append("..")
        for p in remainingPath:
            relativePath.append(p)
        if isinstance(srcFile.page, Md):
            relativePath = [".."] + relativePath

        relativePathStr = os.path.join("", *relativePath)

        return relativePathStr

    def invokePyPage(self, fileNode: FileNode, env: dict[str, Any]) -> None:
        assert fileNode.page is not None
        print(f"{Fore.grey_42}Processing:{Style.reset}", fileNode.fullPath)
        env = env.copy()

        # Enrich with current file:
        env |= {"file": fileNode}

        toProcessFurther: str
        if isinstance(fileNode.page, NonMd) or isinstance(fileNode.page, Md):
            toProcessFurther = fileNode.page.fileContent
        else:
            raise Exception(f"{fileNode} pyPage attribute is invalid.")

        # Inject link
        def link(name: str) -> str:
            return self.link(fileNode, name)

        env |= {"link": link}

        if isinstance(fileNode.page, Page):
            env |= {"lastUpdated": fileNode.page.lastUpdated}
        if isinstance(fileNode.page, Md) and fileNode.page.draftDate is not None:
            env |= {"draftDate": fileNode.page.draftDate}

        # Invoke pypage
        pyPageOutput = pypage(toProcessFurther, env)

        # Perform Markdown processing
        if isinstance(fileNode.page, Md):
            mdResult = Md.processMarkdown(pyPageOutput)
            env.update(mdResult.metadata)
            pyPageOutput = mdResult.html

        if "public" in env:
            if env["public"] is True:
                fileNode.makePublic()
            elif env["public"] is False:
                fileNode.shouldPublish = False

        if isinstance(fileNode.page, Md):
            templateHtml = Content.getTemplateHtml(env)
            # Re-process against `templateHtml` with PyPage:
            pyPageOutput = pypage(templateHtml, env | {"body": pyPageOutput})

        fileNode.pyPageOutput = pyPageOutput

    def process(self) -> None:
        def walk(node: DirNode, env: dict[str, Any]) -> None:
            env = env.copy()

            # Enrich with current dir:
            env |= {"dir": node}

            # Run a __config__.py file, if one exists.
            configEnv = env.copy()
            if config_py_file in (f.fileName for f in node.files):
                print(
                    f"{Fore.dark_red_2}Running:{Style.reset}",
                    os.path.join(node.fullPath, config_py_file),
                )
                exec(readfile(config_py_file), configEnv)
            env |= self.getModuleVars(configEnv)

            # Ordering Note: We must recurse into the subdirectories first.
            for d in node.subDirs:
                with enterDir(d.dirName):
                    walk(d, env)

            # Ordering Note: Files in the current directory must be processed after
            # all subdirectories have been processed so that they have access to
            # information about the subdirectories.
            for f in node.files:
                if f.page is not None:
                    self.invokePyPage(f, env)

        initial_env = self.getBasicHelpers()
        walk(self.rootDir, initial_env)

    @staticmethod
    def getModuleVars(env: Dict[str, Any]) -> Dict[str, Any]:
        return {
            k: v
            for k, v in env.items()
            if (not k.startswith("_") and not isinstance(v, types.ModuleType))
        }

    @staticmethod
    def getBasicHelpers() -> Dict[str, Any]:
        return {"readfile": readfile, "sh": sh}

    @staticmethod
    def splitPath(path: str) -> List[str]:
        head, tail = os.path.split(path)
        if head == "":
            return [path]
        else:
            return Content.splitPath(head) + [tail]

    @staticmethod
    def getTemplateHtml(env: dict[str, Any]) -> str:
        if "template" not in env:
            raise Exception(
                "You must define a `template` var in some ancestral `__config__.py` file."
            )
        template = env["template"]
        if not isinstance(template, str):
            raise Exception("The `template` must be a string.")
        return template


@contextmanager
def enterDir(newDir: str) -> Generator[None, None, None]:
    # https://stackoverflow.com/a/13847807/908430
    oldDir = os.getcwd()
    os.chdir(newDir)
    try:
        yield
    finally:
        os.chdir(oldDir)


def run(args: Args) -> None:
    startTimeNs = time_ns()
    contentDir = args.content
    if not os.path.isdir(contentDir):
        raise Exception(f"The provided path '{contentDir}' is not a directory.")

    with enterDir(args.content):
        rootDir, nameRegistry = fs_crawl()
        print(nameRegistry)
        content = Content(args, rootDir, nameRegistry)
        print("Processing...\n")
        content.process()
        print("\nSuccessfully completed processing.\n")

    print("File Tree:")
    print(displayDir(rootDir))

    generate(args, content)

    elapsedMilliseconds = (time_ns() - startTimeNs) / 10**6
    print("\nTime elapsed: %.2f ms" % elapsedMilliseconds)


def generate(args: Args, content: Content) -> None:
    outputDir = args.output

    def walk(curDir: DirNode) -> None:
        for subDir in curDir.subDirs:
            if subDir.shouldPublish:
                os.mkdir(subDir.dirName)
                with enterDir(subDir.dirName):
                    walk(subDir)

        for fileNode in curDir.files:
            if fileNode.shouldPublish:
                if fileNode.page is not None:
                    assert fileNode.pyPageOutput is not None
                    assert isinstance(fileNode.pyPageOutput, str)

                    if isinstance(fileNode.page, Md):
                        os.mkdir(fileNode.pageName)
                        with enterDir(fileNode.pageName):
                            with open("index.html", "w") as pageHtml:
                                pageHtml.write(fileNode.pyPageOutput)
                    elif isinstance(fileNode.page, NonMd):
                        fileName = fileNode.page.rectifiedFileName
                        if os.path.exists(fileName):
                            raise Exception(
                                f"File {fileName} already exists, and conflicts with {fileNode}."
                            )
                        with open(fileName, "w") as nonMdFile:
                            nonMdFile.write(fileNode.pyPageOutput)

                    else:
                        raise Exception(f"{fileNode} pyPage attribute is invalid.")

                else:
                    if args.copy_assets:
                        shutil.copyfile(fileNode.absoluteFilePath, fileNode.fileName)
                    else:
                        os.symlink(fileNode.absoluteFilePath, fileNode.fileName)

    resetOutputDir(outputDir)

    print("Generating...")

    with enterDir(outputDir):
        walk(content.rootDir)


def resetOutputDir(outputDir: str) -> None:
    if os.path.isfile(outputDir):
        raise Exception("There is a file named %s." % outputDir)
    if os.path.isdir(outputDir):
        print("Deleting directory %s and all of its content...\n" % outputDir)
        shutil.rmtree(outputDir)
    os.mkdir(outputDir)



import re
import os
import yaml
import markdown
import sh  # type: ignore
from collections import defaultdict
from datetime import date, datetime
from subprocess import check_output, CalledProcessError, STDOUT
from typing import (
    Optional,
    List,
    Dict,
    DefaultDict,
    Set,
    Tuple,
    Callable,
    Union,
    NamedTuple,
)
from colored import Style, Fore  # type: ignore

colored_logs = True
config_py_file = "__config__.py"


class FsNode(object):
    def __init__(
        self, parent: Optional["FsNode"], dirPath: str, fileName: Optional[str]
    ) -> None:
        self.parent = parent
        self.fileName: Optional[str] = fileName
        self.dirName: str = os.path.basename(dirPath)
        self.fullPath: str = (
            (os.curdir if dirPath == "" else dirPath)
            if isinstance(self, DirNode) or type(fileName) is not str
            else os.path.join(dirPath, fileName)
        )
        self.shouldPublish: bool = False

    def __repr__(self) -> str:
        return self.colorize(self.fullPath)

    def colorize(self, r: str) -> str:
        if colored_logs:
            if self.shouldPublish:
                r = f"{Style.bold}{Fore.spring_green_2b}{r}{Style.reset}"
        return r

    def setNodeAsPublic(self) -> None:
        self.shouldPublish = True

    def makePublic(self) -> None:
        runOnFsNodeAndAscendantNodes(self, lambda fsNode: fsNode.setNodeAsPublic())


class FileNode(FsNode):
    def __init__(self, parent: Optional[FsNode], dirPath: str, fileName: str) -> None:
        super().__init__(parent, dirPath, fileName)
        self.fileName: str = fileName  # for typing
        split_name = os.path.splitext(fileName)
        self.basename: str = split_name[0]
        self.extension: str = split_name[1]
        self.absoluteFilePath: str = os.path.join(os.getcwd(), self.fullPath)

        self.page: Optional[Union[Md, NonMd]] = None
        self.pageName: str = self.basename  # to be overwritten selectively
        self.pyPageOutput: Optional[str] = None  # to be generated (by pypage)

    def colorize(self, r: str) -> str:
        if colored_logs:
            if self.page is not None and self.shouldPublish:
                r = f"{Fore.spring_green_1}{r}{Style.reset}"
            elif isinstance(self.page, Md):
                r = f"{Fore.purple_4b}{r}{Style.reset}"
        return r

    def __repr__(self) -> str:
        r = f"{self.fileName}"
        r = self.colorize(r)
        r = super().colorize(r)
        return r


class DirNode(FsNode):
    def __init__(
        self,
        parent: Optional[FsNode],
        dirPath: str,
        # shouldIgnore(name: str, isDir: bool) -> bool
        shouldIgnore: Callable[[str, bool], bool],
    ) -> None:
        _, subDirNames, fileNames = next(os.walk(dirPath))
        dirPath = "" if dirPath == os.curdir else dirPath
        super().__init__(parent, dirPath, None)

        self.files: List[FileNode] = [
            FileNode(self, dirPath, fileName)
            for fileName in fileNames
            if not shouldIgnore(fileName, False)
        ]
        self.subDirs: List[DirNode] = [
            DirNode(self, os.path.join(dirPath, subDirName), shouldIgnore)
            for subDirName in subDirNames
            if not shouldIgnore(subDirName, True)
        ]


def displayDir(dirNode: DirNode, indent: int = 0) -> str:
    return (
        (" " * 2 * indent)
        + "%s -> %s\n" % (dirNode, dirNode.files)
        + "".join(displayDir(subDir, indent + 1) for subDir in dirNode.subDirs)
    )


class NameRegistry(object):
    def __init__(self, root: DirNode, skipForRegistry: Callable[[str], bool]) -> None:
        self.allFiles: Dict[str, FileNode] = {}
        self.skipForRegistry = skipForRegistry

        allFilesMulti: DefaultDict[str, Set[FileNode]] = defaultdict(set)

        def record(fileNode: FileNode) -> None:
            allFilesMulti[fileNode.fileName].add(fileNode)

            if fileNode.page is not None:
                allFilesMulti[fileNode.basename].add(fileNode)  # maybe delete this

                if fileNode.basename != fileNode.pageName:
                    allFilesMulti[fileNode.pageName].add(fileNode)

        def walk(node: DirNode) -> None:
            for f in node.files:
                if not self.skipForRegistry(f.fileName):
                    record(f)
            for d in node.subDirs:
                walk(d)

        walk(root)

        # TODO: Handle index pages specially here

        for name, fileNodes in allFilesMulti.items():
            assert len(fileNodes) >= 1
            if len(fileNodes) > 1:
                self.errorOut(name, fileNodes)

            self.allFiles[name] = fileNodes.pop()

    def lookup(self, name: str) -> FileNode:
        if name not in self.allFiles:
            print(
                f"Link error: `{name}` was not found in the name registry."
                f" The {self}"
            )
            raise Exception(f"Link error: {name}")

        return self.allFiles[name]

    @staticmethod
    def errorOut(name: str, fileNodes: Set[FileNode]) -> None:
        raise Exception(
            f"Error: The name '{name}' has multiple matches:\n"
            + "  \n".join(f" {fileNode.fullPath}" for fileNode in fileNodes)
        )

    def __repr__(self) -> str:
        return (
            f"Name Registry:\n  "
            + "\n  ".join(
                f"{k}: {v} {Fore.grey_39}@ {v.fullPath}{Style.reset}"
                for k, v in self.allFiles.items()
            )
            + "\n"
        )


class Page(object):
    def __init__(self, f: FileNode) -> None:
        self.fileContent: str = readfile(f.fullPath)
        self.lastUpdated: datetime = self.getLastUpdated(f.fullPath)

    @staticmethod
    def getLastUpdated(path: str) -> datetime:
        """Get last modified date from: (a) git history, or (b) system modified time."""
        if Page.isGitRepo():
            lastUpdated = Page.getGitFileLastAuthDate(path)
            if lastUpdated is not None:
                return lastUpdated
        return datetime.fromtimestamp(os.path.getmtime(path))

    @staticmethod
    def isGitRepo() -> bool:
        try:
            check_output(["git", "status"], stderr=STDOUT).decode()
            return True
        except CalledProcessError as e:
            return False

    @staticmethod
    def getGitFileLastAuthDate(path: str) -> Optional[datetime]:
        try:
            return datetime.fromisoformat(
                check_output(
                    ["git", "log", "-n", "1", "--pretty=format:%aI", path]
                ).decode()
            )
        except Exception as e:
            return None


class Md(Page):
    def __init__(self, f: FileNode) -> None:
        super().__init__(f)
        self.draftDate: Optional[date] = None

        # Handle file names that start with a date
        dateFragmentLength = len("YYYY-MM-DD-")
        if len(f.basename) > dateFragmentLength:
            dateFragment_ = f.basename[:dateFragmentLength]
            remainingBasename = f.basename[dateFragmentLength:]
            if re.match("[0-9]{4}-[0-9]{2}-[0-9]{2}-$", dateFragment_):
                dateFragment = dateFragment_[:-1]
                self.draftDate = date.fromisoformat(dateFragment)
                f.pageName = remainingBasename

    class Result(NamedTuple):
        metadata: Dict[str, str]
        html: str

    @staticmethod
    def processMarkdown(text: str) -> Result:
        md = markdown.Markdown(extensions=["meta", "codehilite"])
        html: str = md.convert(text)
        yamlFrontMatter: str = ""

        for name, lines in md.Meta.items():  # type: ignore
            yamlFrontMatter += "%s : %s \n" % (name, lines[0])
            for line in lines[1:]:
                yamlFrontMatter += " " * (len(name) + 3) + line + "\n"

        metadata = yaml.safe_load(yamlFrontMatter)
        if metadata is None:
            metadata = dict()
        if not isinstance(metadata, dict):
            raise Exception("Expected yaml.safe_load to return a dict or None.")

        return Md.Result(metadata=metadata, html=html)


class NonMd(Page):
    def __init__(self, f: FileNode, pageName: str, rectifiedFileName: str) -> None:
        super().__init__(f)
        f.pageName = pageName
        self.rectifiedFileName: str = rectifiedFileName

    @staticmethod
    def isNonMdPyPageFile(fileNode: FileNode) -> Optional["NonMd"]:
        """Check if fileNode is a non-Md page (that needs to be processed with pypage).
        If is a non-Md page, we return a NonMd object.
        If it is not, we return None."""
        if ".py." in fileNode.fileName:
            pySubExtPos = fileNode.fileName.find(".py.")
            remainingExt = fileNode.fileName[pySubExtPos:]
            expectedRemainingExt = ".py" + fileNode.extension
            if remainingExt == expectedRemainingExt:
                # The condition above passing indicates this is a NonMd page file.
                realPageName = fileNode.fileName[:pySubExtPos]
                rectifiedFileName = realPageName + fileNode.extension
                return NonMd(fileNode, realPageName, rectifiedFileName)
        return None


def readPages(node: FsNode) -> None:
    if isinstance(node, DirNode):
        d: DirNode = node
        for aDir in node.subDirs:
            readPages(aDir)
        for aFile in node.files:
            readPages(aFile)

        if any(aDir.shouldPublish for aDir in node.subDirs) or any(
            aFile.shouldPublish for aFile in node.files
        ):
            d.shouldPublish = True

    elif isinstance(node, FileNode):
        f: FileNode = node
        if f.extension == ".md":
            f.page = Md(f)
        else:
            f.page = NonMd.isNonMdPyPageFile(f)


def readfile(file_path: str) -> str:
    with open(file_path, "r") as a_file:
        return a_file.read()


def runOnFsNodeAndAscendantNodes(
    startingNode: FsNode, fn: Callable[[FsNode], None]
) -> None:
    def walk(node: FsNode) -> None:
        fn(node)
        if node.parent is not None:
            walk(node.parent)

    walk(startingNode)


def isHidden(name: str) -> bool:
    return name.startswith(".")


def defaultShouldIgnore(name: str, isDir: bool) -> bool:
    if isHidden(name):
        return True
    if name in {"__pycache__"}:
        return True
    basename, fileExt = os.path.splitext(name)
    if fileExt == ".pyc":
        return True
    return False


def defaultSkipForRegistry(name: str) -> bool:
    if name == config_py_file:
        return True
    return False


def fs_crawl(
    # Signature -- shouldIgnore(name: str, isDir: bool) -> bool
    shouldIgnore: Callable[[str, bool], bool] = defaultShouldIgnore,
    skipForRegistry: Callable[[str], bool] = defaultSkipForRegistry,
) -> Tuple[DirNode, NameRegistry]:
    """
    Crawl the current directory. Construct & return an FsNode tree and NameRegistry.
    """
    dirPath: str = os.curdir
    rootDir: DirNode = DirNode(
        None, dirPath, lambda s, b: shouldIgnore(s, b) or defaultShouldIgnore(s, b)
    )

    readPages(rootDir)  # This must occur before NameRegistry creation.

    nameRegistry = NameRegistry(
        rootDir, lambda s: skipForRegistry(s) and defaultSkipForRegistry(s)
    )

    return rootDir, nameRegistry



#!/usr/bin/env python3
import sys

from .engine import run, Args


def main() -> None:
    run(Args().parse_args())


# See: https://chriswarrick.com/blog/2014/09/15/python-apps-the-right-way-entry_points-and-scripts/

if __name__ == "__main__":
    main()



from setuptools import setup

version = "0.7.0"

name = "alteza"

repo_url = "https://github.com/arjun-menon/%s" % name
download_url = "%s/archive/v%s.tar.gz" % (repo_url, version)

setup(
    name=name,
    version=version,
    description="Super-flexible Static Site Generator",
    # Ref: https://packaging.python.org/en/latest/discussions/install-requires-vs-requirements/
    install_requires=[
        # Standard dependencies:
        "pypage >= 2.0.8",
        "markdown >= 3.4.4",
        "pygments >= 2.16.1",
        "pyyaml >= 6.0.1",
        "colored >= 2.2.3",
        "sh >= 2.0.6",
        "typed-argument-parser >= 1.8.1",
        # Dev dependencies:
        "black",
        "pip-tools",
        "mypy ~= 1.5.1",
        "pyre-check ~= 0.9.18",
        "pyright ~= 1.1.327",
        "pytype ~= 2023.9.19",
        "types-PyYAML >= 6.0.12.12",
        "types-Markdown >= 3.4.2.10",
    ],
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url=repo_url,
    download_url=download_url,
    author="Arjun G. Menon",
    author_email="contact@arjungmenon.com",
    keywords=["static site generator", "static sites", "ssg"],
    license="AGPL-3.0-or-later",
    # Ref:https://docs.python.org/3.11/distutils/examples.html
    packages=["alteza"],
    entry_points={
        "console_scripts": [f"{name}=alteza.__main__:main"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Markup :: Markdown",
        "Topic :: Text Processing :: Markup :: HTML",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "License :: OSI Approved :: GNU Affero General Public License v3",
    ],
)
